/*
 * interpose ptrace() to fix Kylix debugger on recent (>2.6.10) kernels
 */

#define _GNU_SOURCE

#include <stdarg.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include <stdio.h>
#include <linux/user.h>
#include <linux/unistd.h>
#include <sys/types.h>
#include <sys/ptrace.h>

//#define DEBUG

typedef long (*ptrace_func)(enum __ptrace_request, ...);

#define TRAP_FLAG	0x100

#define MAX_CHILD_PROCESSES	128

static int pids[MAX_CHILD_PROCESSES];
static int pid_count = 0;

static int is_single_stepped(int pid)
{
	int i;
	for (i = 0; i < pid_count; i++)
		if (pids[i] == pid) return 1;
	return 0;
}

static void set_single_stepped(int pid, int state)
{
	int i;

	for (i = 0; i < pid_count; i++) {
		if (abs(pids[i]) == pid) {
			pids[i] = state ? pid : -pid;
			return;
		}
	}

	if (pid_count == MAX_CHILD_PROCESSES)
	  return;

	pids[i] = state ? pid : -pid;
	pid_count++;
}

#ifdef DEBUG

// gettid declaration
_syscall0(pid_t,gettid)

const char *req_names[] = {
  "TRACEME  ",  // 0,
  "PEEKTEXT ",  // 1,
  "PEEKDATA ",  // 2,
  "PEEKUSER ",  // 3,
  "POKETEXT ",  // 4,
  "POKEDATA ",  // 5,
  "POKEUSER ",  // 6,
  "CONT     ",  // 7,
  "KILL     ",  // 8,
  "SINGLESTEP",  // 9,
  "<INVALID>",	// 10
  "<INVALID>",	// 11
  "GETREGS  ",  // 12,
  "SETREGS  ",  // 13,
  "GETFPREGS",  // 14,
  "SETFPREGS",  // 15,
  "ATTACH   ",  // 16,
  "DETACH   ",  // 17,
  "GETFPXREGS",  // 18,
  "SETFPXREGS",  // 19,
  "<INVALID>",	// 20
  "<INVALID>",	// 21
  "<INVALID>",	// 22
  "<INVALID>",	// 23
  "SYSCALL  ",  // 24
};
#endif

long int ptrace (enum __ptrace_request __request, ...)
{
	static ptrace_func func = NULL;
	struct user_regs_struct *regs;
	va_list va;
	pid_t pid;
	void *addr, *addr2;
	int data;
	long int ret;
	unsigned long old_flags = 0;

	if (!func)
		func = (ptrace_func) dlsym(RTLD_NEXT, "ptrace");

	va_start(va, __request);
	pid = va_arg(va, pid_t);
	addr = va_arg(va, void *);
	data = va_arg(va, int);
	addr2 = va_arg(va, void *);
	va_end(va);

#ifdef DEBUG
	printf("pid %d (%d) param: %s %d %p %d %p    ", getpid(), gettid(), req_names[__request], pid, addr, data, addr2);
#endif

	regs = (struct user_regs_struct *) data;

	if (__request == PTRACE_SETREGS) {
		old_flags = regs->eflags;
		if (regs->eflags & TRAP_FLAG) {
			struct user_regs_struct tmp_regs;
			ret = func(PTRACE_GETREGS, pid, NULL, &tmp_regs);
			regs->eflags = (regs->eflags & ~TRAP_FLAG) | (tmp_regs.eflags & TRAP_FLAG);
#ifdef DEBUG
			if (regs->eflags & TRAP_FLAG) {
				printf("\n  SETREGS with trap flag requested\n");
			} else {
				printf("\n  we are clearing the trap flag\n");
			}
#endif
		}
	}

	ret = func(__request, pid, addr, data, addr2);

	if (__request == PTRACE_SETREGS)
		regs->eflags = old_flags;

	if (__request == PTRACE_GETREGS) {
		if (is_single_stepped(pid)) {
			regs->eflags |= TRAP_FLAG;
			set_single_stepped(pid, 0);
#ifdef DEBUG
			printf("\n  we are setting the trap flag");
#endif
		}
#ifdef DEBUG
		printf("\n  flags %08lx (trap flag %d)\n", regs->eflags, (regs->eflags & TRAP_FLAG) != 0);
#endif
	}

#ifdef DEBUG
	printf("ret: %08lx\n", ret);
#endif

	if (__request == PTRACE_SINGLESTEP)
		set_single_stepped(pid, 1);
//	else if (__request == PTRACE_CONT)
//		set_single_stepped(pid, 0);

	return ret;
}

